import empyrical
import pandas as pd
import numpy as np

from . import plotter


def generate_return_stats(daily_returns, flip_mdd) -> None:
    """
    generates following returns statics for each Ntile:
        - Sharpe
        - Annual Return
        - Annual Vol
        - % Periods Up
        - Max Drawdown (flips for top and bottom bins, excluding middle bin)

        If long_short = Ture:
            - All above calculated on spread and universe
            - Annual Tracking Error
            - Information Ratio
    :param daily_returns: the returns we are calculating stats for
    :param flip_mdd: should max draw down be flipped around the center?
    """
    ntile_funcs = {
        'sharpe': sharpe_ratio,
        'CAGR': simple_returns_CGAR,
        'Vol': annual_volatility,
        'Max Drawdown': lambda x: max_drawdown(x, flip_mdd),
        '% Periods Up': percent_periods_up,
    }

    calculated_stats = [func(daily_returns) for func in ntile_funcs.values()]
    render_me = pd.DataFrame(calculated_stats).transpose()
    plotter.render_heat_table(render_me)


def compute_ntile_stats(name, func, ntile_returns) -> pd.Series:
    """
    apply a function to each column of ntile_returns
    :param name: name of the function
    :param func: the function to apply
    :param ntile_returns: the returns we are applying the function to
    :return: pd.Series, index: Ntile; Name: name;
    """
    return ntile_returns.apply(func, axis=0).rename(name)


def max_drawdown(daily_returns, flip_bottom) -> pd.Series:
    """
    computes the max drawdown for each column
    flips the drawdown from downside to upside for ntiles that should be negative
    :param daily_returns: the returns we are getting the drawdown for
    :param flip_bottom:
    :return: pd.Series, index: Ntile; Values: drawdown
    """
    adj_ret = daily_returns.copy()  # gets rid of setting on copy warning
    num_cols = daily_returns.shape[1]

    if flip_bottom:
        mid_pos = int(round(num_cols / 2 + .5)) - 1
        adj_ret.iloc[:, mid_pos:] = adj_ret.iloc[:, mid_pos:] * -1
        out = compute_ntile_stats('Max Drawdown', empyrical.max_drawdown, adj_ret)

        if num_cols % 2 == 1:  # if even number of columns is odd pad null
            out.iloc[mid_pos] = None

        return out * 100

    return compute_ntile_stats('Max Drawdown', empyrical.max_drawdown, adj_ret) * 100


def percent_periods_up(daily_returns) -> pd.Series:
    """
    computes the percent of periods where return is > 0
    :param daily_returns: the returns we are getting the % of periods up for
    :return: pd.Series, index: Ntile; Values: % periods up
    """
    periods_up = daily_returns.copy()
    periods_up.iloc[:] = np.where(daily_returns.values > 0, 1, 0)
    return (periods_up.sum(axis=0) / periods_up.shape[0]).rename('% Periods Up')


def annual_volatility(daily_returns) -> pd.Series:
    """
    computes the annual volatility of each column
    :param daily_returns: the returns we are getting the vol for
    :return: pd.Series, index: Ntile; Values: annual vol
    """
    return compute_ntile_stats('Annual Vol', empyrical.annual_volatility, daily_returns) * 100


def sharpe_ratio(daily_returns) -> pd.Series:
    """
    computes the sharpe ratio for each
    :param daily_returns: the returns we are getting the sharpe of
    :return:pd.Series, index: Ntile; Values: sharpe
    """
    return compute_ntile_stats('Sharpe', empyrical.sharpe_ratio, daily_returns)


def simple_returns_CGAR(daily_returns) -> pd.Series:
    """
    computes the CAGR form simple returns
    :param daily_returns: make t e
    :return: series with index: cum_returns.columns; values: corresponding average return in percent
    """
    return CAGR(cum_returns(daily_returns))


def CAGR(cum_returns_df: pd.DataFrame) -> pd.Series:
    """
    calculates the geometric average yearly ntile returns from the given cumulative returns
    Assumed the data is in daily format
    :param cum_returns_df: cn be full cum returns or just the
    :return: series with index: cum_returns_df.columns; values: corresponding average return in percent
    """
    return ((cum_returns_df.iloc[-1] ** (1 / (cum_returns_df.shape[0] / 252)) - 1) * 100).rename('CAGR')


def cum_returns(simple_returns: pd.DataFrame) -> pd.DataFrame:
    """
    Calculates the daily returns from the simple returns.
    wraps empyrical.cum_returns
    :param simple_returns: returns used to calculate the cumulative returns
    :return: cumulative returns
    """
    return empyrical.cum_returns(simple_returns, starting_value=1)



import warnings
from typing import Dict, Iterable, Optional

import pandas as pd
import duckdb

from .portals.base_portal import BaseGrouperPortalConstant
from .portals.pricing_portal import PricingPortal
from .tears.base_tear import BaseTear
from .tears.ic_tear import ICHorizonTear, ICTear
from .tears.inspection_tear import InspectionTear
from .tears.tilts_backtest_tear import TiltsBacktestTear
from .tears.turnover_tear import TurnoverTear


class Ntile:
    def __init__(self, pricing_portal: PricingPortal, group_portal: Optional[BaseGrouperPortalConstant] = None):
        """
        :param pricing_portal: the pricing portal which holds pricing data for all assets with factor values
        :param group_portal: group portal which holds grouping information for all assets with factor values
            if this is None then no group statistics will be calculated
        """
        self._pricing_portal: PricingPortal = pricing_portal
        self._group_portal = group_portal

        self._factor_data = None
        self._ntile_matrix = None
        self._formatted_returns = None

    def _input_checks(self, factor_series) -> None:
        """
        checks the factor series to ensure it meet requirements to run a tearsheet

        Requirements:
            1) series must have MultiIndex with 2 levels
            2) First level must be of type pd.Period
            3) PricingPortal must have data for all Period dates in the series
            4) There can only be one observations for a single asset on a single day

        :param factor_series: the series we are checking
        :return: None
        :raise ValueError: if one of the requirements are not met
        """

        # checking for series with multi index, possibly also check types for multi index
        if not isinstance(factor_series.index, pd.MultiIndex) or factor_series.index.nlevels != 2:
            raise ValueError('Factor input must have MultiIndex of period, id')

        # ensure the index level zero is date
        if not isinstance(factor_series.index.get_level_values(0), pd.PeriodIndex):
            raise ValueError('Factor input must have MultiIndex with the first level being a period '
                             f'current factor dtype is {type(factor_series.index.get_level_values(0))}')

        # we will check id when looking for overlapping portal names
        no_pricing_for = set(factor_series.index.get_level_values(1)).difference(
            self._pricing_portal.assets)
        if len(no_pricing_for) != 0:
            # raise ValueError(f'PricingPortal does not have data for: {no_pricing_for}')
            warnings.warn(f'PricingPortal does not have data for: {no_pricing_for}')

        # make sure pricing portal dates match up with factor
        overlapping_periods = set(factor_series.index.get_level_values(0).drop_duplicates()).intersection(
            self._pricing_portal.periods)
        if len(overlapping_periods) == 0:
            raise ValueError('No overlap between PricingPortal dates and factor dates')
        if len(overlapping_periods) < 100:
            warnings.warn(f'Only {len(overlapping_periods)} common dates between PricingPortal dates and factor')

        # check for multiple observations on a single day for a single asset
        if factor_series.index.duplicated().any():
            raise ValueError('Multiple factor observations on single day for a single asset')

    def _set_ntiles_and_returns(self, factor_data: pd.Series, ntiles: int):
        """
        Sets self._formatted_returns and  self._formatted_ntile
        :param factor_data: the factor data
        :param ntiles: amount of ntiles
        :return: None
        """
        self._ntile_factor_sql(factor_data, ntiles)
        self._align_ntiles_pricing()

        # can see what % of the dataframe is null here
        self._make_null_summary(factor_data)

    def _align_ntiles_pricing(self) -> None:
        """
        ensures ntiled matrix and daily returns matrix have the same column and row order
        sets self._formatted_returns and self._ntile_matrix
        :return: None
        """
        ntile_factor = self._factor_data['ntile'].unstack()
        daily_returns = self._pricing_portal.delta_data

        factor_date = ntile_factor.index.get_level_values('date')
        self._formatted_returns = daily_returns[(daily_returns.index >= factor_date.min()) &
                                                (daily_returns.index <= factor_date.max())]

        # reindexing the ntiles data so that you have pricing and ntiles matching up
        self._ntile_matrix = ntile_factor.reindex_like(self._formatted_returns)

    def _make_null_summary(self, raw_factor_data) -> None:
        """
        making a summary of how much factor data we matched to pricing data
        :param raw_factor_data: the raw unstacked factor data
        """
        length_og_factor_data = len(raw_factor_data)
        # seeing what % of factor data is missing
        num_na_data_points = raw_factor_data.isnull().sum()
        pct_na_data_points = num_na_data_points / length_og_factor_data

        # amount of data droped because of non aligned factor and returns dates:
        # above should be non null length of ntiles before reindexing
        # non null length of ntiles after indexing
        number_of_finite_ntiles = length_og_factor_data - num_na_data_points
        binary_if_ntile_data = self._ntile_matrix.notnull()
        number_of_finite_ntiles_no_overlap_returns = number_of_finite_ntiles - binary_if_ntile_data.sum().sum()
        pct_missing_ntile_no_overlap = number_of_finite_ntiles_no_overlap_returns / number_of_finite_ntiles

        # amount of data we dont have returns for given we have overlapping pricing and factor
        # should ffill ntile by holdign period since we need return data holding_period days out
        binary_if_return_data = self._formatted_returns.notnull()
        # should forward fill by holding period to make sure we have pricing for when we will be holding the stock
        missing_from_no_returns_given_overlap = (number_of_finite_ntiles
                                                 - (binary_if_ntile_data * binary_if_return_data).sum().sum())
        pct_missing_data_no_returns_given_overlap = missing_from_no_returns_given_overlap / number_of_finite_ntiles

        # total number of unusable factor data points due to null or no maped returns
        num_bad = (num_na_data_points
                   + number_of_finite_ntiles_no_overlap_returns
                   + missing_from_no_returns_given_overlap
                   )

        pct_bad = num_bad / length_og_factor_data

        print(f"Unusable Factor Data:           {(round(pct_bad, 4)) * 100}%")
        print(f"NA Factor Values:               {(round(pct_na_data_points, 4)) * 100}%")
        print(f"No Overlapping Returns:         {(round(pct_missing_ntile_no_overlap, 4)) * 100}%")
        print(f"Missing Returns Given Overlap:  {(round(pct_missing_data_no_returns_given_overlap, 4)) * 100}%")

    def _ntile_factor(self, factor: pd.Series, ntiles: int) -> None:
        """
        This is slow replaced by
        Universe relative Quantiles of a factor by day _ntile_factor_sql

        pd.DataFrame of ntiled factor
            index: (pd.Period, _asset_id)
            Columns: (factor, ntile)
            Values: (factor value, Ntile corresponding to factor value)

        :param factor: same var as ntile_return_tearsheet
        :param ntiles: same var as ntile_return_tearsheet
        """
        # add a filter for if a day has less than 20% factor data then just put bin as -1 for all assets
        # unstack the frame, percentile rank each row, divide whole matrix buy 1/ntiles, take the floor of every number
        factor = factor[~factor.isnull()].to_frame('factor')

        try:
            factor['ntile'] = factor.groupby('date').transform(
                lambda date_data: ntiles - pd.qcut(date_data, ntiles, labels=False)
            ).sort_index()
        except Exception as e:
            print('Hit error while binning data. Need to push the histogram')
            print('Your data is mighty sus we can\'t Ntile it. This is normally due to bad data')

            # forcing a histogram out
            import matplotlib.pyplot as plt
            factor.groupby('date').count().plot()
            plt.show()

            raise e

        self._factor_data = factor

    def _ntile_factor_sql(self, factor: pd.Series, ntiles: int) -> None:
        """
        Universe relative Quantiles of a factor by day
        Around 100X faster than pandas groupby qcut

        pd.DataFrame of ntiled factor
            index: (pd.Period, _asset_id)
            Columns: (factor, ntile)
            Values: (factor value, Ntile corresponding to factor value)

        :param factor: same var as ntile_return_tearsheet
        :param ntiles: same var as ntile_return_tearsheet
        """
        factor = factor.to_frame('factor').reset_index()
        factor['date'] = factor['date'].dt.to_timestamp()

        sql_quantile = f"""SELECT *, NTILE({ntiles}) OVER(PARTITION BY date ORDER BY factor.factor DESC) as ntile
                            FROM factor
                            WHERE factor.factor IS NOT NULL"""
        con = duckdb.connect(':memory:')
        factor = con.execute(sql_quantile).df()
        factor['date'] = factor['date'].dt.to_period(freq='D')
        factor = factor.set_index(['date', 'id'])

        self._factor_data = factor

    #
    # Start up methods
    #
    def _prep_for_run(self, factor: pd.Series, ntiles: int) -> None:
        """
        prepares the ntiles class to run a tear sheet
        :param factor: factor for tear sheet
        :param ntiles: num ntiles for sheet
        :return: None
        """
        # checking to see if we have series or data frame
        if isinstance(factor, pd.DataFrame):
            if factor.shape[1] > 1:  # there is a df passed with multible columns
                raise ValueError('There are multiple columns in the passed DataFrame')

            factor_series = factor.iloc[:, 0]
        else:
            factor_series = factor.copy()

        self._input_checks(factor_series)

        factor_series.index.names = ['date', 'id']
        self.kick_tears(factor_series, ntiles)

        self._print_start_end_dates()

    def _print_start_end_dates(self):
        """
        prints the start and end date of the backtest
        """
        date = self._factor_data.index.get_level_values(0)
        print(f'\nStart Date: {date.min()}')
        print(f'End Date:   {date.max()}\n')

    def kick_tears(self, factor_series: pd.Series, ntiles: int) -> None:
        """
        Clears the object of all factor and tear data.
        Reruns Ntiling of factor
        :param factor_series: the user passed factor
        :param ntiles: the number of ntiles
        :return: None
        """
        self._clear()
        self._set_ntiles_and_returns(factor_series, ntiles)

    def _clear(self) -> None:
        """
        clears all data points in the object except the pricing portal
        :return: None
        """
        self._factor_data = None
        self._ntile_matrix = None
        self._formatted_returns = None

    @staticmethod
    def _run(tears: Dict[str, BaseTear]) -> None:
        """
        Runs all tear sheets that are set in the class
        :return: None
        """
        for tear in tears.values():
            tear.compute_plot()

    #
    # Tear Sheets Below
    #
    def full_tear(self, factor: pd.Series, ntiles: int, holding_period: int, long_short: bool = True,
                  market_neutral=True, show_uni=False, show_ntile_tilts=False) -> Dict[str, BaseTear]:
        """
        Creates basic visualizations of the factor data distribution by ntile and how complete the data is
        Creates a fan chart of cumulative returns for the given factor values.
        Creates a IC time series for the factor value and the forward returns
        Createa a turnover sheet showing how often the factor data will turn over

        The in the cumulative return plot, each value represents the cumulative return up to that days close.
        Returns are not shifted each value represents portfolios value on the close of that day.

        A set of weights is generated for each day based off factor quantile.
        The portfolio is rebalanced daily, each days 1/holding_period of the portfolio is rebalanced.
        All positions are equally weighted.

        :param factor: The factor values being tested.
            index: (pd.Period, _asset_id)
            values: (factor_value)
        :param holding_period: How long we want to hold positions for, represents days
        :param ntiles: amount of bins we are testing (1 is high factor value n is low value)
        :param long_short: show we compute the spread between ntiles: (1 - n)
        :param market_neutral: subtract out the universe returns from the ntile returns?
        :return: plots showing the return profile of the factor
        :param show_uni: Should universe return be shown in the spread plot?
        :param show_ntile_tilts: should we show each ntiles tilts?
        """
        self._prep_for_run(factor, ntiles)
        tears = {'inspection_tear': InspectionTear(factor_data=self._factor_data),
                 'backtest_tear': TiltsBacktestTear(ntile_matrix=self._ntile_matrix,
                                                    daily_returns=self._formatted_returns, ntiles=ntiles,
                                                    holding_period=holding_period, long_short=long_short,
                                                    market_neutral=market_neutral,
                                                    show_uni=show_uni, factor_data=self._factor_data,
                                                    group_portal=self._group_portal,
                                                    show_ntile_tilts=show_ntile_tilts),
                 'ic_tear': ICTear(factor_data=self._factor_data, daily_returns=self._formatted_returns,
                                   holding_period=holding_period),
                 'turnover_tear': TurnoverTear(factor_data=self._factor_data, holding_period=holding_period)}
        self._run(tears)
        return tears

    def ntile_backtest_tear(self, factor: pd.Series, ntiles: int, holding_period: int, long_short: bool = True,
                            market_neutral=True, show_uni=False, show_ntile_tilts=False) -> Dict[str, BaseTear]:
        """
        Creates a fan chart of cumulative returns for the given factor values.
        The factor values are ntile'd into ntiles number of bins

        The in the cumulative return plot, each value represents the cumulative return up to that days close.
        Returns are not shifted each value represents portfolios value on the close of that day.

        A set of weights is generated for each day based off factor quantile.
        The portfolio is rebalanced daily, each days 1/holding_period of the portfolio is rebalanced.
        All positions are equally weighted.

        :param factor: The factor values being tested.
            index: (pd.Period, _asset_id)
            values: (factor_value)
        :param holding_period: How long we want to hold positions for, represents days
        :param ntiles: amount of bins we are testing (1 is high factor value n is low value)
        :param long_short: show we compute the spread between ntiles: (1 - n)
        :param market_neutral: subtract out the universe returns from the ntile returns?
        :return: plots showing the return profile of the factor
        :param show_uni: Should universe return be shown in the spread plot?
        :param show_ntile_tilts: should we show each ntiles tilts?
        """
        self._prep_for_run(factor, ntiles)
        tears = {'backtest_tear':
                     TiltsBacktestTear(ntile_matrix=self._ntile_matrix, daily_returns=self._formatted_returns,
                                       ntiles=ntiles, holding_period=holding_period, long_short=long_short,
                                       market_neutral=market_neutral, show_uni=show_uni, factor_data=self._factor_data,
                                       group_portal=self._group_portal, show_ntile_tilts=show_ntile_tilts)
                 }
        self._run(tears)
        return tears

    def ntile_inspection_tear(self, factor: pd.Series, ntiles: int) -> Dict[str, BaseTear]:
        """
        creates visuals showing the factor data over time
        only calculates IC for when the asset is in the universe
        :param factor: The factor values being tested.
            index: (pd.Period, _asset_id)
            values: (factor_value)
        :param ntiles: the number of ntiles
        :return: Dict of InspectionTear
        """
        self._prep_for_run(factor, ntiles)
        tears = {'inspection_tear': InspectionTear(factor_data=self._factor_data)}
        self._run(tears)
        return tears

    def ntile_ic_tear(self, factor: pd.Series, holding_period: int) -> Dict[str, BaseTear]:
        """
        creates visuals showing the ic over time
        :param factor: The factor values being tested.
            index: (pd.Period, _asset_id)
            values: (factor_value)
        :param holding_period: How long we want to hold positions for, represents days
        :return: Dict of ICTear
        """
        self._prep_for_run(factor, 1)
        tears = {'ic_tear': ICTear(factor_data=self._factor_data, daily_returns=self._formatted_returns,
                                   holding_period=holding_period)}
        self._run(tears)
        return tears

    def ntile_turnover_tear(self, factor: pd.Series, ntiles: int, holding_period: int) -> Dict[str, BaseTear]:
        """
        Creates visuals showing the turnover over time
        :param factor: The factor values being tested.
           index: (pd.Period, _asset_id)
           values: (factor_value)
        :param ntiles: the number of ntiles
        :param holding_period: How long we want to hold positions for, represents days
        :return: Dict of TurnoverTear
        """
        self._prep_for_run(factor, ntiles)
        tears = {'turnover_tear': TurnoverTear(factor_data=self._factor_data, holding_period=holding_period)}
        self._run(tears)
        return tears

    def ntile_ic_horizon(self, factor: pd.Series, intervals: Iterable[int], show_individual: bool = False) -> \
            Dict[str, BaseTear]:
        """
        Shows the curve of the information coefficient over various holding periods

        :param factor: The factor values being tested.
           index: (pd.Period, _asset_id)
           values: (factor_value)
        :param intervals: an iterable that contains the holding periods we would like to make the IC frontier for
        :param show_individual: should each individual IC time series be show for every interval
        :return: Dict of ICHorizonTear
        """
        self._prep_for_run(factor, 1)
        tears = {
            'ic_horizon_tear': ICHorizonTear(factor_data=self._factor_data, daily_returns=self._formatted_returns,
                                             intervals=intervals, show_individual=show_individual)}
        self._run(tears)
        return tears



import pandas as pd

from ntiles import Ntile, PricingPortal

uni = pd.read_csv('/Users/alex/Documents/Notebook/improved_momentum/data/index/crsp_sp500/mapped_index.csv')
assets = uni.lpermno.astype(str).unique().tolist()
pricing_portal = PricingPortal(assets, '2015', '2016')
tile = Ntile(pricing_portal)
alpha = pricing_portal.delta_data.shift(1).stack()
tile.full_tear(alpha, ntiles=5, holding_period=2)



from abc import ABC
from typing import Iterable

import pandas as pd

from ntiles import plotter
from ntiles.tears.base_tear import BaseTear
from ntiles import utils


class ICTear(BaseTear, ABC):
    """
    Computes IC from the given factor and returns

    Currently will only measure IC for days a company is in the universe
    Example: AAPl is in the univere on 1/10 but not in universe on 11/10 if we have greater than 10 day holding period
        that asset wint count in the IC calculation
    """

    def __init__(self, factor_data: pd.DataFrame, daily_returns: pd.DataFrame, holding_period: int):
        """
        :param factor_data: factor data to look at must be from Ntiles
        :param daily_returns: daily returns we are calculating the IC on must be from Ntiles
        :param holding_period: Holding period we are calculating IC for
        """
        super().__init__()
        self.factor_data = factor_data
        self.daily_returns = daily_returns
        self.holding_period = holding_period

        self.daily_ic = None
        self.ic_stats = None

    #
    # Calculation
    #
    def compute(self) -> None:
        """
        master function for computing the IC
        :return: None
        """""
        self.compute_daily_ic()
        self.calculate_ic_table()

    def compute_daily_ic(self) -> None:
        """
        calculates and sets the daily IC for the holding period
        :return: None
        """
        self.factor_data.index.names = ['date', 'id']

        # slicing off factor values we dont have forward return data for
        factor_unstacked = self.factor_data['factor'].unstack()#.iloc[:-self.holding_period]
        forward_returns = self.compute_forward_returns().reindex_like(factor_unstacked)

        ic_array = utils.correlation_2d(factor_unstacked.to_numpy(), forward_returns.to_numpy())
        self.daily_ic = pd.Series(ic_array, index=forward_returns.index)

    def compute_forward_returns(self) -> pd.DataFrame:
        """
        Calculates self.holding_period forward returns from daily returns
        :return: index: date; columns: asset; values: self.holding_period forward returns
        """
        # must mad extra day due to cumprod making first date nan
        daily_ret = self.daily_returns  # utils.pad_extra_day(self.daily_returns, 0)
        return daily_ret.add(1).cumprod().pct_change(self.holding_period).shift(-self.holding_period)

    def calculate_ic_table(self) -> None:
        """
        calculates summary stats for the IC data
        :return: None, sets self.ic_stats
        """
        mean_ic = self.daily_ic.mean()
        std_ic = self.daily_ic.std()
        stats = {
            'IC Mean': mean_ic,
            'IC Median': self.daily_ic.median(),
            'IC Std': std_ic,
            'Risk Adjusted IC': mean_ic / std_ic,
            'IC Skew': self.daily_ic.skew()
        }

        self.ic_stats = pd.Series(stats).round(3).to_frame(f'{self.holding_period}D').transpose()

    #
    # Plotting
    #
    def plot(self) -> None:
        """
        plots the IC data in self.daily_ic
        :return: None
        """
        print('Information Coefficient')
        plotter.render_table(self.ic_stats)
        plotter.plot_timeseries_ic(self.daily_ic, self.holding_period)
        # plotter.plot_ic_qq(self.daily_ic)
        # plotter.plot_ic_hist(self.daily_ic)

    #
    # To clipboard functions
    #
    def ic_to_clipboard(self) -> None:
        """
        writes ic to the clipboard
        :return: None
        """
        self.daily_ic.to_clipboard()


class ICHorizonTear(BaseTear, ABC):
    """
    Computes the IC horizon tear
    Will give insight into optimal holding periods for the factor
    """

    def __init__(self, factor_data: pd.DataFrame, daily_returns: pd.DataFrame, intervals: Iterable[int],
                 show_individual):
        """
        :param factor_data: The factor values being tested, must be from Ntiles
        :param daily_returns: matrix of returns from Ntiles
        :param intervals: an iterable that contains the holding periods we would like to make the IC frontier for
        """
        super().__init__()
        self._factor_data = factor_data
        self._daily_returns = daily_returns
        self._intervals = sorted(list(intervals))
        self._show_individual = show_individual

        self.tears = {}
        self._ic_horizon = None

    def compute(self) -> None:
        """
        runs a IC tear for all the periods we want to test over
        """
        for interval in self._intervals:
            self.tears[interval] = ICTear(self._factor_data, self._daily_returns, interval)
            self.tears[interval].compute()

        self._ic_horizon = pd.concat([tear.ic_stats for tear in self.tears.values()])

    def plot(self) -> None:
        """
        plots the IC frontier and the Time series IC
        """
        plotter.plot_ic_horizon(self._ic_horizon.drop(['IC Skew'], axis=1))
        plotter.render_table(self._ic_horizon)
        if self._show_individual:
            for ic_tear in self.tears.values():
                ic_tear.plot()



from abc import ABC
from typing import List

import pandas as pd

try:
    from equity_db import MongoAPI, ReadDB
except ImportError:
    pass

from ntiles.portals.base_portal import BaseGrouperPortalConstant


class SectorPortal(BaseGrouperPortalConstant, ABC):
    def __init__(self, passed_assets: List[str], asset_id: str = 'lpermno', db: str = 'equity',
                 collection: str = 'crsp'):
        """
        :param asset_id: the assets to get the sector data for
        :param asset_id: what is the id of the asset, must be recognised by equity_db
        :param db: name of the db
        :param collection: name of the collection
        """
        super().__init__(passed_assets, 'GIC Sector')
        self._passed_assets = passed_assets
        self._asset_id = asset_id
        self._db = db
        self._collection = collection

        self._sectors = None
        self._set_sectors()

    @property
    def group_information(self) -> pd.Series:
        """
        gets the gic _sectors for the give assets
        :return: DataFrame of GIC _sectors for the given assets
        """
        if self._sectors is not None:
            return self._sectors

        self._set_sectors()
        return self._sectors

    @property
    def group_mapping(self):
        """
        :return: dict mapping for the group
        """
        return self.group_information.to_dict()

    def _set_sectors(self) -> None:
        """
        Sets the _sectors in the class
        :return: None
        """
        reader = ReadDB(MongoAPI(db=self._db, collection=self._collection))
        query = reader.get_asset_data(self._passed_assets, search_by=self._asset_id, fields=['gsector'])
        self._sectors = query.df['gsector']
        self._sectors.index = self._sectors.index.astype(str)

    @property
    def assets(self) -> List[int]:
        return self._sectors.reset_index().lpermno.astype(int).tolist()



import warnings
from abc import ABC

import numpy as np
import pandas as pd

from ntiles.tears.base_tear import BaseTear
from ntiles import plotter, stats, utils


class BacktestTear(BaseTear, ABC):
    """
    Computes returns and stats from the given factor and pricing data

    Upgrades:
        Have cash account for when security gets delisted and we own it
        One day holding period

    """

    def __init__(self, ntile_matrix: pd.DataFrame, daily_returns: pd.DataFrame, ntiles, holding_period: int,
                 long_short: bool, market_neutral: bool, show_uni: bool):
        """
        :param ntile_matrix: unstacked and formatted ntiles prepared by Ntiles
        :param daily_returns: unstacked and formatted daily returns from Ntiles
        :param holding_period: How long we want to hold positions for, represents days
        :param ntiles: amount of bins we are testing (1 is high factor value n is low value)
        :param long_short: show we compute the spread between ntiles: (1 - n)
        :param market_neutral: subtract out the universe returns from the ntile returns?
        :param show_uni: suhould universe return be shown in the spread plot?
        """

        super().__init__()

        self.ntile_matrix = ntile_matrix
        self.daily_returns = daily_returns
        self.ntiles = ntiles
        self.holding_period = holding_period
        self.long_short = long_short
        self.market_neutral = market_neutral
        self.show_uni = show_uni

        self.daily_weights = {}
        self.weighted_returns = {}
        self._daily_tile_returns = None

    def compute(self) -> None:
        """
        method to run the backtest
        """
        self.kick_backtest()

    def plot(self):
        """
        method to plot the data for the backtest
        """
        self.kick_visuals()

    #
    # Vectorized Ntile Backtest
    #
    def kick_backtest(self):
        """
        Calculates the daily returns of each ntile
        Saves the daily returns in self._daily_tile_returns
            index: pd.Period
            columns: Ntile: {ntile}
            Values: Daily close ntile returns on corresponding day
        :return: None
        """

        daily_ntile_returns = self._get_ntile_returns_helper()

        if self.long_short:
            daily_ntile_returns[f'1 vs {self.ntiles}'] = (daily_ntile_returns.iloc[:, 0] -
                                                          daily_ntile_returns.loc[:, f'Ntile: {self.ntiles}']) / 2

            if self.ntiles > 3:
                daily_ntile_returns[f'2 vs {self.ntiles - 1}'] = (daily_ntile_returns.iloc[:, 1] -
                                                                  daily_ntile_returns.loc[:,
                                                                  f'Ntile: {self.ntiles - 1}']) / 2

        self._daily_tile_returns = daily_ntile_returns

    def _get_ntile_returns_helper(self) -> pd.DataFrame:
        """
        Helper to get the returns for each ntile on each day
        :return: data frame index: pd.period; columns: Ntile; values: daily returns
        """
        np_ntile_matrix = self.ntile_matrix.to_numpy()
        np_asset_returns_matrix = self.daily_returns.to_numpy()

        out = {}
        for ntile in range(1, self.ntiles + 1):
            out[f'Ntile: {ntile}'] = self._compute_daily_ntile_returns(np_ntile_matrix, np_asset_returns_matrix, ntile,
                                                                       self.holding_period)

        universe_ntile_matrix = np.where(np.isfinite(np_ntile_matrix), 1, np.nan)[self.holding_period - 1:]
        universe_returns_matrix = np_asset_returns_matrix[self.holding_period - 1:]

        out['universe'] = self._compute_daily_ntile_returns(universe_ntile_matrix, universe_returns_matrix, 1, 1)

        if self.holding_period != 1:
            out = pd.DataFrame(out, index=self.ntile_matrix.index[self.holding_period - 2:])
        else:
            raise ValueError('One day holding period is currently not supported!')
            #print()
            #out = pd.DataFrame(out, index=self.ntile_matrix.index)

        if self.market_neutral:
            # subtracting out universe returns
            ntile_cols = utils.get_ntile_cols(out)
            out.loc[:, ntile_cols] = out.loc[:, ntile_cols].subtract(out['universe'], axis=0)

        if not self.show_uni:
            out.drop('universe', axis=1, inplace=True)

        return out

    def _compute_daily_ntile_returns(self, ntile_matrix: np.array, asset_returns_matrix: np.array, ntile: int,
                                     holding_period: int) -> np.array:
        """
        Computes the daily returns for a ntile
        :param ntile_matrix: the matrix of ntiles
        :param asset_returns_matrix: the matrix for returns
        :param ntile: the amount of ntiles we have computed
        :param holding_period: how long we are holding the assets for
        :return: 1d np.array of the daily return for the ntile
        """

        #
        # Calculating the asset weight per day
        #
        weight_per_day = 1 / np.count_nonzero(ntile_matrix == ntile, axis=1) / holding_period
        if (weight_per_day > .05).any():
            warnings.warn('We have a asset with daily weight over 5%\nLimiting weight at 5%')
            weight_per_day = np.minimum(weight_per_day, np.full(weight_per_day.shape, .05))

        raw_daily_weights = np.where(ntile_matrix == ntile, np.expand_dims(weight_per_day, axis=1), 0)
        daily_weights = utils.rolling_sum(raw_daily_weights, holding_period)

        weighted_asset_returns = daily_weights * asset_returns_matrix[holding_period - 1:, :]
        daily_returns = np.insert(np.sum(weighted_asset_returns, axis=1), 0, 0)

        self.record_backtest_components(ntile, daily_weights, weighted_asset_returns)

        return daily_returns

    def record_backtest_components(self, ntile, daily_weights, weighted_asset_returns):
        """
        records the components to compute the backtest for a specific ntile
        :param ntile: the ntile the data is for
        :param daily_weights: the weights of each asset on the corresponding day
        :param weighted_asset_returns: the weighted returns of each asset
        :return: None
        """
        self.daily_weights[f'Ntile: {ntile}'] = \
            pd.DataFrame(daily_weights, index=self.daily_returns.index[self.holding_period - 1:],
                         columns=self.daily_returns.columns)

        self.weighted_returns[f'Ntile: {ntile}'] = \
            pd.DataFrame(weighted_asset_returns, index=self.daily_returns.index[self.holding_period - 1:],
                         columns=self.daily_returns.columns)

    #
    # Visuals
    #
    def kick_visuals(self) -> None:
        """
        controls displaying visuals to the user
        :return: None
        """
        print('Ntile Backtest')
        cum_ret = stats.cum_returns(self._daily_tile_returns)

        # ntile stats
        ntile_cols = utils.get_ntile_cols(self._daily_tile_returns)
        ntile_daily_ret = self._daily_tile_returns[ntile_cols]
        ntile_cum_ret = cum_ret[ntile_cols]
        avg_annual_ret = stats.CAGR(ntile_cum_ret)
        # ntile plotting
        stats.generate_return_stats(ntile_daily_ret, self.market_neutral)
        plotter.ntile_return_plot(ntile_cum_ret, f'Ntile Returns {self.holding_period} Day Holding Period')
        plotter.ntile_annual_return_bars(avg_annual_ret, self.holding_period)

        if self.long_short:
            # spread stats
            spread_cols = utils.get_non_ntile_cols(self._daily_tile_returns)
            long_short_frame = self._daily_tile_returns[spread_cols]
            # spread plotting
            stats.generate_return_stats(long_short_frame, False)
            plotter.ntile_return_plot(cum_ret[spread_cols],
                                      f'Long Short Returns {self.holding_period} Day Holding Period')

    #
    # Data methods
    #
    def cum_ret_to_clipboard(self) -> None:
        """
        write cumulative returns to clipboard
        :return: None
        """
        stats.cum_returns(self._daily_tile_returns).to_clipboard()



from typing import Union, List

from abc import ABC

import duckdb
import pandas as pd
import numpy as np

from .base_tear import BaseTear
from .. import plotter, utils


class TurnoverTear(BaseTear, ABC):
    """
    Shows the turnover for a factor
    """

    def __init__(self, factor_data: pd.DataFrame, holding_period: Union[int, List[int]]):
        super().__init__()
        self._factor_data = factor_data
        self._holding_period = holding_period

        self._auto_corr = None
        self._turnover = None
        self._summary_stats = dict()

    def compute(self) -> None:
        """
        calculates the data for the tear
        """

        self.calculate_autocorrelation()
        self.calculate_turnover()

        self.calculate_summary_stats()

    def plot(self) -> None:
        """
        plots the tear
        """
        self.plot_turnover()

    def calculate_autocorrelation(self) -> None:
        """
        Calculates the auto correlation of the factor with a lag of self._holding_period

        calculates the autocorrelation of n and n - holding period
        """
        factor_unstacked = self._factor_data['factor'].unstack()
        auto_corr_arr = utils.correlation_2d(factor_unstacked.to_numpy(),
                                             factor_unstacked.shift(self._holding_period).to_numpy())

        self._auto_corr = pd.Series(auto_corr_arr, index=factor_unstacked.index)

    def calculate_turnover(self):
        """
        Calculates the turnover of the top and bottom bin with a lag of self._holding_period

        calculates the turnover of n and n - holding period
        """
        # getting frame of only the top and bottom bin
        max_ntile = self._factor_data['ntile'].max()
        turnover_frame = self._factor_data[['ntile']][self._factor_data['ntile'].isin([1, max_ntile])]
        turnover_frame['ntile_shifted'] = turnover_frame['ntile'].unstack().shift(self._holding_period).stack()
        turnover_frame['changed'] = turnover_frame['ntile'] != turnover_frame['ntile_shifted']

        # fd = self._factor_data[['ntile']].reset_index()
        # fd['date'] = fd['date'].dt.to_timestamp()
        # max_ntile = fd['ntile'].max()
        # turnover_sql = f"""SELECT "date", "ntile",
        #                         "ntile" != lag("ntile", {self._holding_period}) OVER (PARTITION BY id ORDER BY "date") as "changed"
        #                     FROM fd
        #                     WHERE "ntile" in (1, {max_ntile})"""
        #
        # con = duckdb.connect(':memory:')
        # turnover_frame = con.execute(turnover_sql).df()
        # con.close()

        final_turnover = turnover_frame.groupby(['date', 'ntile']).changed.agg(sum=sum, count=len)

        self._turnover = (final_turnover['sum'] / final_turnover['count']).unstack()

    def calculate_summary_stats(self) -> None:
        """
        sets the summary stats for the autocorelation and the turnover
        """
        self._summary_stats['auto'] = self._auto_corr.agg(
            {'Mean AC': np.mean, 'Median AC': np.median, 'Std AC': np.std}).round(3).to_frame(
            f'{self._holding_period}D').transpose()

        self._summary_stats['turnover'] = self._turnover.stack().groupby('ntile').agg(
            **{'Mean Turnover': np.mean, 'Median Turnover': np.median, 'Std Turnover': np.std}).round(3)

    def plot_turnover(self) -> None:
        """
        plots the time series data in self.auto_corr
        """
        print('Autocorrelation')
        plotter.render_table(self._summary_stats['auto'])
        plotter.plot_auto_corr(self._auto_corr, self._holding_period)

        print('Turnover')
        plotter.render_table(self._summary_stats['turnover'])
        plotter.plot_turnover(self._turnover, self._holding_period)



from abc import ABC
from typing import List, Optional

import pandas as pd

try:
    from equity_db import MongoAPI, ReadDB
except ImportError:
    pass

from ntiles.portals.base_portal import BaseDeltaPortal

"""
No Longer used. See toolbox for the pricing portal.
"""


class PricingPortal(BaseDeltaPortal, ABC):
    """
    Object to query and cache pricing data
    """

    def __init__(self, assets: List[str], start: str, end: str, search_by: str = 'lpermno',
                 pricing_field: str = 'prccd', adjustor_field: str = 'ajexdi', db: str = 'equity',
                 collection: str = 'crsp', trading_calender='NYSE'):
        """
        :param assets: The assets to get data for
        :param start: start of period to query
        :param end: end of period to query
        :param pricing_field: what field to use for pricing data
        :param adjustor_field: What field to use for adjusting the pricing data
        :param db: the data base to use
        :param collection: the _collection to query
        :param trading_calender: the trading calendar to use to verify dates
        """
        super().__init__(assets, pd.Period(start, 'D'), pd.Period(end, 'D'))

        self._pricing_field = pricing_field

        self._adjusted_pricing: Optional[pd.DataFrame] = None
        self._period_delta: Optional[pd.DataFrame] = None
        self._query_adjusted_pricing(db, collection, assets, start, end, search_by, pricing_field, adjustor_field,
                                     trading_calender)

    @property
    def delta_data(self):
        """
        :return: unstacked daily asset returns
            col: _asset_id; index: pd.period; values: daily asset returns
        """
        if self._period_delta is None:
            self._period_delta = self.raw_data.unstack().pct_change(1).iloc[1:].fillna(0)

        return self._period_delta

    @property
    def raw_data(self) -> pd.Series:
        """
        adjustments to AssetQuery:
            1) Turns date column from pd.Timestamp into pd.Period
            2) Turns the lpermno into an int
            3) Adjusts the pricing_field: pricing_field / adjustor_field
        :return: Series of the adjusted pricing data indexed by date, lpermno
        """
        if self._adjusted_pricing is None:
            raise ValueError('adjusted pricing is not set')

        return self._adjusted_pricing[self._pricing_field]

    @property
    def assets(self) -> List[int]:
        """
        casting to int due to _db problem must fix
        :return: The id's of assets we have pricing data for
        """
        return self._adjusted_pricing.index.get_level_values('id').astype(int).unique().tolist()

    @property
    def periods(self) -> List[pd.Period]:
        """
        :return: the unique periods for which we have pricing data
        """
        return self._adjusted_pricing.index.get_level_values('date').unique().tolist()

    def _query_adjusted_pricing(self, db, collection, assets, start, end, search_by, pricing_field, adjustor_field,
                                trading_calender) -> None:
        """
        Makes query the pricing data
        Performs adjustments defined in self.daily_pricing
        Then caches the adjusted pricing in self._adjusted_pricing
        self._adjusted_pricing columns: self._pricing_field, self._adjustor_field, Index: date, lpermno
        :return: None, mutates self._adjusted_pricing to contain adjusted pricing
        """
        # querying pricing
        reader = ReadDB(MongoAPI(db, collection))
        query_df = reader.get_asset_data(assets, search_by=search_by, start=pd.Timestamp(start), end=pd.Timestamp(end),
                                         fields=[pricing_field, adjustor_field])
        pricing_df = query_df.set_calendar(trading_calender).df.reset_index()

        # possibly send command to close mongo to free up memory

        # adjusting data frame
        pricing_df[pricing_field] = pricing_df[pricing_field] / pricing_df[adjustor_field]
        pricing_df['date'] = pricing_df['date'].dt.to_period(freq='D')

        # currently code is requiring lpermno input wont work with tickers need to fix _db
        pricing_df['id'] = pricing_df['lpermno'].astype(int)
        pricing_df = pricing_df.set_index(['date', 'id'])

        self._adjusted_pricing = pricing_df

        self._query_summary(assets)  # this can be cleaner

    def _query_summary(self, assets):
        """
        prints a summary of query and tells you what id's were not able to be found in the query
        :return: None
        """
        query_assets = self._adjusted_pricing.index.get_level_values(1).astype(str).unique().tolist()
        not_found_assets = set(assets) - set(query_assets)
        if len(not_found_assets) == 0:
            print('All assets retrieved in query!')
        else:
            print(f'Unable to find {len(not_found_assets)} assets: {not_found_assets}')



from abc import ABC
from typing import Iterable, List, Union

import pandas as pd

from toolbox import QueryConstructor
from .base_portal import BaseGrouperPortalConstant


class SectorPortal(BaseGrouperPortalConstant, ABC):
    def __init__(self, assets: Union[Iterable, str], search_by: str = 'permno', field='gsector', con=None,
                 start_date=None, end_date=None, ):
        """
        :param assets: the assets or universe to get the sector data for
        :param search_by: what is the id of the asset
        :param field: name of field we want to get
        """
        super().__init__(assets, 'GIC Sector')
        self._search_by = search_by
        self._field = field
        self._con = con
        self._start_date = start_date
        self._end_date = end_date

        self._group = None
        self._set_sectors()

    @property
    def group_information(self) -> pd.Series:
        """
        gets the gic _sectors for the give assets
        :return: DataFrame of GIC _sectors for the given assets
        """
        return self._group

    @property
    def group_mapping(self):
        """
        :return: dict mapping for the group
        """
        return self.group_information.to_dict()

    def _set_sectors(self) -> None:
        """
        Sets the _sectors in the class
        :return: None
        """
        self._group = (QueryConstructor(self._con)
                       .query_no_date_table(table='link.crsp_cstat_link', fields=[self._field, 'lpermno as permno'],
                                            assets=self._assets, search_by=self._search_by, start_date=self._start_date,
                                            end_date=self._end_date)
                       .df)[self._field].fillna(-1)

    @property
    def assets(self) -> List[int]:
        return self._group.index.tolist()



#%%

import pandas as pd
from toolbox.constitutes.constitute_adjustment import ConstituteAdjustment
from ntiles.tears import NTile
from ntiles.portals.pricing_portal import PricingPortal


#%%
internet_uni = pd.read_csv('/Users/alex/Documents/Notebook/internet_uni/data/index/internet_uni.csv',
                           usecols=['lpermno', 'int_from', 'int_thru']).rename(
    {'int_from': 'from', 'int_thru': 'thru'}, axis=1)

start = pd.Timestamp('2017')
end = pd.Timestamp('2021')

ca = ConstituteAdjustment('lpermno')
ca.add_index_info(index_constitutes=internet_uni, start_date=start, end_date=end,
                  date_format='%Y-%m-%d')
#%%
traffic_df = pd.read_csv('/Users/alex/Desktop/traffic_history.csv').rename({'datadate': 'date'},
                                                                           axis=1).drop_duplicates(
    subset=['date', 'lpermno'])
traffic_df = traffic_df[traffic_df['weburl'] != 'apple.com']
traffic_df['rank'] = traffic_df['rank'].astype(float)
traffic_df['pvp_million'] = traffic_df['pvp_million'].astype(float)
traffic_df['pvp_user'] = traffic_df['pvp_user'].astype(float)
traffic_df['reach_per_million'] = traffic_df['reach_per_million'].astype(float)
#traffic_df['date'] = pd.to_datetime(traffic_df['date'], format='%Y-%m-%d').dt.to_period(freq='D')
traffic_df = traffic_df.set_index(['date', 'lpermno']).unstack().shift(1).stack().reset_index()

factor_df = ca.adjust_data_for_membership(data=traffic_df, contents='factor', date_format='%Y-%m-%d', )
factor_df = factor_df.reset_index()
factor_df['date'] = factor_df['date'].dt.to_period('D')
factor_df = factor_df.set_index(['date', 'lpermno'])

alpha = factor_df['pvp_million'] / factor_df['pvp_user']


#%%
assets = internet_uni.lpermno.astype(str).unique().tolist()
portal = PricingPortal(assets, start, end)
tile = NTile(portal)


tile.ntile_return_tearsheet(alpha, 5, 5, False)



from abc import ABC, abstractmethod
from typing import List, Union

import pandas as pd


class BasePortal(ABC):
    """
    Base class for the data portal object
    """

    def __init__(self, assets: List[Union[str, int]]):
        """
        :param assets: the assets we are querying for
        """
        self._assets = assets

    @property
    @abstractmethod
    def assets(self) -> List[Union[str, int]]:
        """
        returns the assets property
        """
        return self._assets


class BaseTimeSeriesPortal(BasePortal):
    """portal for time series data"""

    def __init__(self, assets: List[Union[str, int]], start: pd.Period, end: pd.Period):
        """
        :param assets: the assets we are querying for
        :param start: start date for the query
        :param end: end date for the query
        """
        super().__init__(assets)
        self._start = start
        self._end = end

    @property
    @abstractmethod
    def periods(self) -> List[pd.Period]:
        """
        :return: the unique periods for which we have data
        """
        pass


class BaseRawPortal(BaseTimeSeriesPortal, ABC):
    def __init__(self, assets: List[Union[str, int]], start: pd.Period, end: pd.Period):
        """
        :param assets: the assets we are querying for
        :param start: start date for the query
        :param end: end date for the query
        """
        super().__init__(assets, start, end)

    @property
    @abstractmethod
    def raw_data(self) -> pd.DataFrame:
        """
        returns the raw data held by the portal
        :return: Index: Id, pd.Period; Columns: 'data'; Values: data
        """
        pass


class BaseDeltaPortal(BaseTimeSeriesPortal, ABC):
    """
    a portal which fetches and calculates the raw data long with delta or percent delta of a variable.
    Useful for fetching and calculating returns
    """

    @property
    @abstractmethod
    def delta_data(self):
        """
        returns the delta of the data held by the portal
        :return: Index: Id, pd.Period; Columns: 'delta'; Values: data
        """
        pass


class BaseGrouperPortalConstant(BasePortal, ABC):
    """
    A portal which fetches grouping data
    """

    def __init__(self, assets: List[Union[str, int]], group_name: str):
        """
        :param assets: the assets we are querying for
        :param group_name: the name of the grouping
        """
        super().__init__(assets)
        self.group_name = group_name

    @property
    def name(self):
        """
        :return: Name of group
        """
        return self.group_name

    @property
    @abstractmethod
    def group_information(self) -> pd.Series:
        """
        Holds group information from the portal
        :return: Index: Id; Columns: 'group'; Values: group
        """
        pass

    @property
    @abstractmethod
    def group_mapping(self):
        """
        :return: dict mapping for the group
        """
        pass


class BaseGrouperPortalTimeSeries(BaseTimeSeriesPortal, ABC):
    """
    a portal which returns grouping information over a time period
    """

    def __init__(self, assets: List[Union[str, int]], start: pd.Period, end: pd.Period, group_name: str):
        """
        :param assets: the assets we are querying for
        :param start: start date for the query
        :param end: end date for the query
        :param group_name: the name of the grouping
        """
        super().__init__(assets, start, end)
        self.group_name = group_name

    @property
    def name(self):
        """
        :return: Name of group
        """
        return self.group_name

    @property
    @abstractmethod
    def periods(self) -> List[pd.Period]:
        """
        :return: the unique periods for which we have data
        """
        pass

    @property
    @abstractmethod
    def group_information(self) -> pd.DataFrame:
        """
        Holds a timeseries of group information from the portal
        :return: Index: Id, pd.Period; Columns: 'group'; Values: group
        """
        pass



import pandas as pd
import numpy as np
import numba as nb
import duckdb
from typing import Optional, List, Union


def subset_frame(frame: pd.DataFrame, columns: Optional[List[str]]):
    """
    subsets the given data frame by the given columns
    if the columns are none then the whole frame is returned
    :param frame: the dataframe to subset
    :param columns: the columns we are going to subset by, if none then nothing is done
    :return: Given frame subset by the given columns
    """
    if columns:
        return frame[columns]
    return frame


def get_ntile_cols(frame: pd.DataFrame) -> List[str]:
    """
    :param frame: data frame to get columns of
    :return: all columns in the frame that contain 'Ntile'
    """
    return [col for col in frame.columns if 'Ntile' in col]


def get_non_ntile_cols(frame: pd.DataFrame) -> List[str]:
    """
    :param frame: data frame to get columns of
    :return: all columns in the frame that dont contain 'Ntile'
    """
    return [col for col in frame.columns if 'Ntile' not in col]


def make_nan_inf_summary(df: Union[pd.DataFrame, pd.Series], max_loss: float, print_good: bool = True) -> pd.DataFrame:
    """
    makes a summary fot the the amount of nan and infinity values in the given data frame
    will throw a ValueError if the percent of nan and inf is greater than the given threshold
    prints a summary of the nan's and inf of there are any
    :param df: the data frame we are checking
    :param max_loss: max decimal percent of nan and inf we are allowing the df to contain
    :param print_good: should we print the output if we dropped less then the threshold?
    :return: pandas data frame with the nan and inf dropped
    """
    df_numpy = df.to_numpy()
    nan_array = np.isnan(df_numpy)
    finite_array = np.logical_or(np.isinf(df_numpy), np.isneginf(df_numpy))

    if nan_array.any() or (not finite_array.all()):
        factor_length = len(df)
        amount_nan = nan_array.sum()
        amount_inf = finite_array.sum()
        total_percent_dropped = (amount_nan + amount_inf) / factor_length

        outString = f'Dropped {round(total_percent_dropped * 100, 2)}% of data. ' \
                    f'{round((amount_nan / factor_length) * 100, 2)}% due to nan, ' \
                    f'{round((amount_inf / factor_length) * 100, 2)}% of inf values. Threshold: {max_loss * 100}%\n'

        if total_percent_dropped > max_loss:
            raise ValueError('Exceeded Nan Infinity Threshold. ' + outString)

        # print out string as a summary
        if print_good:
            print(outString)

        # dropping the nans and the infinity values
        df = df.replace([np.inf, -np.inf], np.nan).dropna()

    elif print_good:
        print('Dropped 0% of data')

    return df


def rolling_sum(a, n):
    """
    rolling sum, column wise
    :param a: array to roll and sum
    :param n: length of rolling window
    :return: a[n:, :] of rolling sum
    """
    if n == 1:
        return a

    cum_sum = np.cumsum(a, axis=0)
    cum_sum[n:, :] = cum_sum[n:, :] - cum_sum[:-n, :]
    return cum_sum[n - 1:, :]


@nb.njit(parallel=True)
def correlation_2d(factor: np.array, returns: np.array) -> np.array:
    """
    calculates a timeseries of correlation for the given factor and forward returns
    factor and returns must have EXACTLY the same structure and order of assets/days
    think of each row as a group and we calculate the correlation by groups

    :param factor: 2d np.array, each row represents factor values for different assets on same day
    :param returns: 2d np.array, each row represents forward returns for different assets on same day
    :return:1d np.array representing time series of factor values
    """
    if factor.shape != returns.shape:
        raise ValueError('Factor and returns dont represent same information')

    num_rows = factor.shape[0]
    out = np.empty(shape=num_rows)

    for i in nb.prange(num_rows):
        finite_mask = np.isfinite(factor[i]) & np.isfinite(returns[i])
        out[i] = np.corrcoef(factor[i][finite_mask], returns[i][finite_mask])[0][1]

    return out


def pad_extra_day(matrix_df: pd.DataFrame, pad_value: any) -> pd.DataFrame:
    """
    pads a unstacked frame with a single extra row are the start of the data frame
    :param matrix_df: df to pad, index: pd.Period, columns: any, values: any
    :param pad_value: constant value to insert into a row
    :return: matrix_df with a padded value
    """
    out = matrix_df.copy()
    new_period = (out.index.min().to_timestamp() - pd.DateOffset(1)).to_period('D')
    out.loc[new_period, :] = np.full(shape=out.shape[1], fill_value=pad_value)
    return out.sort_index()  # can make this function better without a sort


def remove_cat_index(frame: Union[pd.Series, pd.DataFrame]) -> Union[pd.Series, pd.DataFrame]:
    """
    if the frame has a categorical index it will remove it
    :return: frame with the categorical index removed
    """
    if frame.index.is_categorical():
        frame.index = frame.index.astype(str)

    return frame


def convert_date_to_period(frame: Union[pd.DataFrame, pd.Series], freq: str = 'D', **kwargs) -> Union[
    pd.DataFrame, pd.Series]:
    """
    converts the date column to a period if the date column is of type timestamp
    if the 'date' column is a period then nothing will be changed
    date can be in the index or columns

    :param frame: the frame containing the date column
    :param freq: the freq for the period
    :return: thr same frame that was passed but 'date' is a partiod.
    """
    index_names = list(frame.index.names)
    frame = frame.reset_index()

    if 'date' in frame.columns:
        frame['date'] = frame['date'].dt.to_period(freq)
        frame.set_index(index_names)
        return frame

    raise ValueError('"date" not found in data frame')


def ntile(factor: pd.Series, ntiles: int, ) -> pd.Series:
    """
    Universe relative Quantiles of a factor by day
    Around 100X faster than pandas groupby qcut

    pd.DataFrame of ntiled factor
        index: (pd.Period, _asset_id)
        Columns: (factor, ntile)
        Values: (factor value, Ntile corresponding to factor value)

    :param factor: same var as ntile_return_tearsheet
    :param ntiles: same var as ntile_return_tearsheet
    """
    factor = factor.to_frame('factor').reset_index()
    index_names = factor.columns.tolist()
    index_names.remove('factor')

    date_is_period = isinstance(factor.date.dtype, pd.core.dtypes.dtypes.PeriodDtype)
    if date_is_period:
        factor['date'] = factor['date'].dt.to_timestamp()

    sql_quantile = f"""SELECT *, NTILE({ntiles}) OVER(PARTITION BY date ORDER BY factor.factor DESC) as ntile
                            FROM factor
                            WHERE factor.factor IS NOT NULL"""
    con = duckdb.connect(':memory:')
    factor_ntile = con.execute(sql_quantile).df()
    con.close()

    if date_is_period:
        factor_ntile['date'] = factor_ntile['date'].dt.to_period(freq='D')

    factor_ntile = factor_ntile.set_index(index_names)
    return factor_ntile



from abc import ABC

from ntiles.tears.base_tear import BaseTear
from ntiles import plotter


class InspectionTear(BaseTear, ABC):
    """
    creates a data inspection sheet
    """

    def __init__(self, factor_data):
        """
        :param factor_data: factor_data from Ntiles
        """
        super().__init__()
        self._factor_data = factor_data

    def compute(self) -> None:
        """
        kicks off the tearsheet
        :return: None
        """
        self.make_summary()

    def plot(self) -> None:
        """
        plots the tearsheet
        """
        self.summary_plots()

    def make_summary(self) -> None:
        """
        calculates the summary statics for the factor by Ntile
        """
        quantile_stats = self._factor_data.groupby('ntile').agg(['median', 'std', 'min', 'max', 'count']).factor
        quantile_stats['count %'] = quantile_stats['count'] / quantile_stats['count'].sum() * 100

        # aesthetics
        quantile_stats = quantile_stats.round(2)
        quantile_stats.columns = [col.title() for col in quantile_stats.columns]
        quantile_stats.index.name = 'Ntile:'

        plotter.render_table(quantile_stats, 'Quantiles Statistics')

    def summary_plots(self) -> None:
        """
        plots the the summary of the factor
        """
        no_index_factor_data = self._factor_data.reset_index().dropna()
        date_agg = no_index_factor_data.groupby('date')
        date_ntile_agg = no_index_factor_data.groupby(['date', 'ntile'])

        plotter.plot_inspection_data(date_agg.factor.count(), 'Universe Count Of Factor Per Day', 'Count')
        plotter.plot_inspection_data(date_ntile_agg.factor.count().unstack(), 'Ntile Count of Factor Per Day', 'Count')
        plotter.plot_inspection_data(date_ntile_agg.factor.median().unstack(), 'Median Factor Value by Ntile', 'Median',
                                     2)



import pandas as pd
import numpy as np

import matplotlib as mpl
import matplotlib.pyplot as plt
from IPython.core.display import display

RETURN_COLOR_MAP = mpl.cm.get_cmap('jet')
TILTS_COLOR_MAP = mpl.cm.get_cmap('tab20')
IC_COLOR_MAP = mpl.cm.get_cmap('tab10')

LARGE_FIGSIZE = 20, 10
MEDIUM_FIGSIZE = 15, 8


def ntile_return_plot(cum_ntile_returns: pd.DataFrame, title: str):
    """
    generates cumulative return plot for a ntiles returns series
    if cols are empty list returns None
    :param cum_ntile_returns: cumulative returns we want to plot
    :param title: title of the plot
    :return: matplotlib axis with the return plot on it
    """

    fig, ax = plt.subplots(1, 1, figsize=LARGE_FIGSIZE)

    cum_ntile_returns.plot(lw=2, ax=ax, cmap=RETURN_COLOR_MAP)
    ax.set(ylabel='Log Cumulative Returns', title=title, xlabel='',
           yscale='symlog')

    ax.legend(loc="center left", bbox_to_anchor=(1, .5))
    ax.set_yscale('log', base=2)
    ax.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.2f'))
    ax.axhline(1, linestyle='-', color='black', lw=1)
    fig.autofmt_xdate()

    plt.show()
    return ax


def ntile_annual_return_bars(avg_annual_ret: pd.Series, period: int):
    """
    generates a box plot of the yearly CAGR for each ntile
    :return: matplotlib axis
    """
    num_ntiles = len(avg_annual_ret)

    _, ax = plt.subplots(1, 1, figsize=MEDIUM_FIGSIZE)
    ax.set(ylabel='% Return', title=f'Annual Return, {period} Day Holding period', xlabel='')

    colors = [RETURN_COLOR_MAP(i) for i in np.linspace(0, 1, num_ntiles)]
    ax.bar(avg_annual_ret.index, avg_annual_ret.to_numpy(), color=colors)
    ax.axhline(0, linestyle='-', color='black', lw=1)

    plt.show()
    return ax


def plot_inspection_data(table: pd.DataFrame, title: str, ylabel: str, decimals: int = 0) -> None:
    """
    plots the inspection data for inspection tear sheets
    :param table: the table to plot
    :param title: the title for the plot
    :param ylabel: y label for plot
    :param decimals: amount of decimals to display on the Y axis
    :return: None
    """

    fig, ax = plt.subplots(1, 1, figsize=MEDIUM_FIGSIZE)
    ax.set(title=title, ylabel=ylabel)
    table.plot(lw=2, ax=ax, cmap=RETURN_COLOR_MAP)
    ax.legend(loc="center left", bbox_to_anchor=(1, .5))
    # ax.xaxis.set_major_formatter(mpl.dates.DateFormatter('%m-%Y'))
    ax.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter(f'%.{decimals}f'))
    fig.autofmt_xdate()

    if isinstance(table, pd.Series):
        ax.get_legend().remove()

    plt.show()


def plot_tilts(frame: pd.DataFrame, ntile: str, group_name: str, ax=None):
    """
    Plots the timeseries group tilts for a single ntile
    :param frame: frame containing the tilts per day, columns: group, index: pd.Period
    :param ntile: the Ntile we are plotting for
    :param group_name: the name of the group
    :param ax: axis to plot on
    :return: None
    """
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=MEDIUM_FIGSIZE)

    ax.set(title=f'{ntile}, {group_name}'.title(), ylabel='Weight In Ntile')
    frame.plot(lw=2, ax=ax, cmap=TILTS_COLOR_MAP, legend=None)
    ax.axhline(0, linestyle='-', color='black', lw=1)
    # ax.xaxis.set_major_formatter(mpl.dates.DateFormatter('%m-%Y'))
    ax.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter(f'%.2f'))
    plt.show()


def plot_tilt_hist(series, ntile: str, group_name: str, extra_space: bool = True):
    """
    Plots the histogram group tilts for a single ntile
    :param series: frame containing the avg tilts, columns: group, index: pd.Period
    :param ntile: the Ntile we are plotting for
    :param group_name: the name of the group
    :return: None
    """
    if extra_space:
        fig, ax = plt.subplots(1, 2, figsize=LARGE_FIGSIZE)
    else:
        _, ax = plt.subplots(1, 1, figsize=(4.5, 4.5))

    title = 'Weight Relative to Universe' if 'Ntile' in group_name else 'Group Exposure'
    plotter_frame = series.to_frame('weight')
    plotter_frame['colors'] = [TILTS_COLOR_MAP(i) for i in np.linspace(0, 1, len(series))]
    plotter_frame = plotter_frame.sort_values('weight')

    ax[0].barh(plotter_frame.index.tolist(), plotter_frame['weight'].tolist(), align='center',
               color=plotter_frame['colors'].tolist())
    ax[0].set(title=f'{ntile}, {group_name}'.title(), ylabel='Group', xlabel=title)
    ax[0].axvline(0, linestyle='-', color='black', lw=1)

    if extra_space:
        return ax[1]

    plt.show()


def plot_timeseries_ic(ic_series: pd.Series, holding_period: int):
    """
    plots the daily time series IC
    :param ic_series: series of IC to plot index: pd.Period
    :param holding_period: how long the holding period is for the IC
    :return: None
    """
    ic_frame = ic_series.to_frame('IC')
    ic_frame['1 Month Avg IC'] = ic_frame.rolling(21).mean()

    fig, ax = plt.subplots(1, 1, figsize=MEDIUM_FIGSIZE)
    ic_frame.plot(ax=ax, title=f'IC {holding_period} Day Holding Period')
    ax.get_lines()[1].set_linewidth(3)
    ax.axhline(0, linestyle='-', color='black', lw=1)
    fig.autofmt_xdate()
    plt.show()


def plot_auto_corr(ac_series: pd.Series, holding_period: int) -> None:
    """
    plots the daily time series IC
    :param ac_series: series of auto corr to plot index: pd.Period
    :param holding_period: how long the holding period is for the IC
    :return: None
    """
    fig, ax = plt.subplots(1, 1, figsize=MEDIUM_FIGSIZE)
    ac_series.plot(ax=ax, title=f'Autocorrelation {holding_period} Day Holding Period')
    ax.axhline(ac_series.median(), linestyle=(0, (5, 10)), color='black', lw=1)
    fig.autofmt_xdate()
    plt.show()


def plot_turnover(turn_frame: pd.Series, holding_period: int) -> None:
    """
    plots the daily time series IC
    :param turn_frame: dataframe of turnover to plot index: pd.Period
    :param holding_period: how long the holding period is for the IC
    :return: None
    """
    fig, ax = plt.subplots(1, 1, figsize=MEDIUM_FIGSIZE)
    colors = [RETURN_COLOR_MAP(i) for i in np.linspace(0, 1, turn_frame.columns.max())]

    for col in turn_frame.columns:
        ax.plot(turn_frame.index.to_timestamp(), turn_frame[col], color=colors[col - 1], label=f'Ntile: {col}')
        ax.axhline(turn_frame[col].median(), linestyle=(0, (5, 10)), color=colors[col - 1], lw=5)

    ax.set(ylabel='% Turnover', title=f'Turnover {holding_period} Day Holding Period', xlabel='')
    ax.legend(loc="center left", bbox_to_anchor=(1, .5))
    fig.autofmt_xdate()
    plt.show()


def plot_ic_horizon(horizon_frame: pd.DataFrame):
    ax_tuple = plt.subplots(2, 2, figsize=LARGE_FIGSIZE)[1].flatten()
    colors = [IC_COLOR_MAP(i) for i in np.linspace(0, 1, 4)]

    for i in range(horizon_frame.shape[1]):
        plot_me = horizon_frame.iloc[:, i]
        plot_me.plot(ax=ax_tuple[i], color=colors[i], title=plot_me.name)
    plt.show()


def render_heat_table(frame: pd.DataFrame) -> None:
    """
    renders a dataframe as a heatmap
    :param frame: the frame to render
    :return: None
    """
    cm = mpl.cm.get_cmap('RdYlGn')
    styled = frame.style.background_gradient(cmap=cm, axis=0).format('{:.2f}').set_properties(
        **{'text-align': 'center'})
    render_table(styled)


def render_table(table: pd.DataFrame, output: str = None) -> None:
    """
    displays a table to the user
    :param table: the table to display
    :param output: the output we should render
    :return: None
    """
    if output:
        print(output)
    display(table)



from distutils.core import setup

from setuptools import find_packages

setup(
    name='ntiles',
    version='0.1.4.1',
    packages=find_packages(),
    license='Apache License 2.0',
    description='Vectorized quantile backtester.',
    url='https://github.com/Alexd14/ntiles-backtester',
    download_url='https://github.com/Alexd14/ntiles-backtester/archive/refs/tags/v0.1.4.1.zip',
    keywords=['factor', 'backtesting', 'alphalens', 'vectorized backtesting', 'equity trading'],
    install_requires=[
        'numba',
        'pandas',
        'numpy',
        'matplotlib',
        'empyrical',
        'factor_toolbox',
        # 'equity-db'
        ],
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



from abc import ABC, abstractmethod


class BaseTear(ABC):
    """
    The base class for all tearsheets
    """

    def __init__(self):
        """
        empty constructor
        """

    def compute_plot(self) -> None:
        """
        method which calculates stats and plots the data for the tearsheet
        :return: None
        """
        self.compute()
        self.plot()

    @abstractmethod
    def compute(self) -> None:
        """
        method which calculates stats for the tearsheet
        :return: None
        """
        pass

    @abstractmethod
    def plot(self) -> None:
        """
        method which plots data for the tearsheet
        :return: None
        """
        pass



from abc import ABC
from typing import Iterable, Union

import pandas as pd
import numpy as np
from .base_portal import BaseDeltaPortal

from toolbox import QueryConstructor


class PricingPortal(BaseDeltaPortal, ABC):
    """
    pulls pricing from database
    """

    def __init__(self, assets: Union[Iterable, str], search_by: str, start_date: str, end_date: str,
                 field: str = 'prc', table: str = 'CRSP.sd', con=None):
        super().__init__(assets, pd.Period(start_date), min(pd.Timestamp(end_date), pd.Timestamp('today')).to_period('D'))
        self._search_by = search_by
        self._field = field
        self._table = table
        self._con = con

        self._pricing = None
        self._get_pricing()

    @property
    def assets(self):
        return self._pricing.columns.tolist()

    @property
    def delta_data(self):
        """
        returns the delta of the data held by the portal
        :return: Index: Id, pd.Period; Columns: 'delta'; Values: data
        """
        return self._pricing

    @property
    def periods(self):
        return self._pricing.index.drop_duplicates().to_list()

    def _get_pricing(self):
        df = (QueryConstructor(sql_con=self._con)
              .query_timeseries_table(self._table, assets=self._assets,
                                      start_date=str(self._start), end_date=str(self._end), search_by=self._search_by,
                                      fields=[self._field])
              .distinct()
              .set_calendar('NYSE')
              .order_by('date')
              .df)

        self._pricing = df[self._field].unstack().pct_change(1).iloc[1:]. \
            fillna(0).replace([np.inf, -np.inf], 0).clip(-.75, 1.5)



from abc import ABC
from typing import Optional

import pandas as pd

from .backtest_tear import BacktestTear
from .. import plotter
from .. import utils

from ..portals.base_portal import BaseGrouperPortalConstant


class TiltsBacktestTear(BacktestTear, ABC):
    """
    generates a tear sheet which shows the sector exposures of a strategy
    Must be run after the backtest tear
    """

    def __init__(self, ntile_matrix: pd.DataFrame, daily_returns: pd.DataFrame, ntiles, holding_period: int,
                 long_short: bool, market_neutral: bool, show_uni: bool, factor_data: pd.DataFrame,
                 group_portal: Optional[BaseGrouperPortalConstant], show_ntile_tilts: bool):
        """
        :param ntile_matrix: unstacked and formatted ntiles prepared by Ntiles
        :param daily_returns: unstacked and formatted daily returns from Ntiles
        :param holding_period: How long we want to hold positions for, represents days
        :param ntiles: amount of bins we are testing (1 is high factor value n is low value)
        :param long_short: show we compute the spread between ntiles: (1 - n)
        :param market_neutral: subtract out the universe returns from the ntile returns?
        :param show_uni: should universe return be shown in the spread plot?
        :param factor_data: the factor data from Ntiles
        :param group_portal: the group portal holding the groups. If this is None then the exposures will not be shown
        :param show_ntile_tilts: Should we show the exposures for each individual ntile?
        """

        super().__init__(ntile_matrix, daily_returns, ntiles, holding_period, long_short, market_neutral, show_uni)
        self._factor_data = factor_data
        self._group_portal = group_portal
        self._show_ntile_tilts = show_ntile_tilts

        self._daily_group_weights = {}
        self._full_group_tilt_avg = {}

    def compute(self) -> None:
        """
        master function for the tear sheet
        :return: None
        """
        super().compute()

        if (self._group_portal is not None) and (self._show_ntile_tilts or self.long_short):
            self.compute_tilts()

    def plot(self) -> None:
        """
        plots the tear sheet
        """
        super().plot()
        if (self._group_portal is not None) and (self._show_ntile_tilts or self.long_short):
            self.make_plots()

    def compute_tilts(self):
        """
        computes the daily tilt data for each group
        :return: None
        """
        self.compute_group_weights()
        if self.long_short:
            self.calculate_long_short_tilts()

    def compute_group_weights(self):
        """
        computes the weights by group for each ntile
        currently computes data but work because need a time series data adjusted for index constitutes
        have to use self.factor_data
        :return: None
        """
        group_info = self._group_portal.group_information
        center_weight = group_info.groupby(group_info).count() / group_info.shape[0]
        center_weight = utils.remove_cat_index(center_weight)

        if self._show_ntile_tilts:
            ntile_keys = self.daily_weights.keys()
        else:
            ntile_keys = [min(self.daily_weights.keys()), max(self.daily_weights.keys())]

        new_col = self.daily_weights[ntile_keys[0]].columns.astype(str).map(self._group_portal.group_mapping)

        for ntile in ntile_keys:
            frame = self.daily_weights[ntile]
            frame.columns = new_col
            frame = self.daily_weights[ntile].stack().to_frame('weight')
            frame.index.names = ['date', 'group']

            weights_unstacked = frame.groupby(['date', 'group']).sum().sub(center_weight, level=1, axis=0).unstack()
            weights_unstacked.columns = weights_unstacked.columns.droplevel(0)

            self._daily_group_weights[ntile] = weights_unstacked
            self._full_group_tilt_avg[ntile] = (frame.groupby('group').sum().weight
                                                / frame.index.levels[0].unique().shape[0]
                                                - center_weight)

    def calculate_long_short_tilts(self):
        """
        calculates the time series tilts for the long short portfolio
        :return: None
        """
        ntile_n = max(self._daily_group_weights.keys())
        self._daily_group_weights['Long Short'] = (self._daily_group_weights['Ntile: 1']
                                                   - self._daily_group_weights[ntile_n])
        self._full_group_tilt_avg['Long Short'] = self._daily_group_weights['Long Short'].stack().groupby(
            'group').mean()

    def make_plots(self):
        print('Weights By Group')
        for ntile in self._daily_group_weights.keys():
            if 'Long Short' == ntile and not self.long_short:
                continue
            if 'Ntile' in ntile and not self._show_ntile_tilts:
                continue
            ax = plotter.plot_tilt_hist(self._full_group_tilt_avg[ntile], ntile, self._group_portal.name)
            plotter.plot_tilts(self._daily_group_weights[ntile], ntile, self._group_portal.name, ax)
